import 'package:flex/expose_flutter.dart';

/// Extensions to assist with color ANSI escape codes.
extension AnsiColor on Color {
  /// The foreground ansi espace params for this [Color].
  String get ansiParamsForeground => '38;2;$red;$green;$blue';

  /// The background ansi espace params for this [Color].
  String get ansiParamsBackground => '48;2;$red;$green;$blue';

  /// Returns [string] wrapped with [ansiParamsForeground] and [defaultForeground] or [ansiParamsBackground] and [defaultBackground], depending on [bg].
  ///
  /// If [alpha] is 0, the returned [String] will instead be **prefixed** with the [defaultForeground] or [defaultBackground] escape codes.
  ///
  /// Usage:
  /// ```dart
  /// final textToColor = 'some error happened';
  /// // define our color; in this case red.
  /// const color = Color(0xff0000);
  /// // wrap our text with the appropriate ANSI sequences.
  /// final coloredText = color.ansiWrap(textToColor);
  /// writeln(coloredText);
  /// ```
  String ansiWrap(String string, {bool bg = false}) {
    if (bg) {
      if (alpha == 0) {
        return '${build([defaultBackground])}$string';
      }
      return '${build([ansiParamsBackground])}$string${build([defaultBackground])}';
    } else {
      if (alpha == 0) {
        return '${build([defaultForeground])}$string';
      }
      return '${build([ansiParamsForeground])}$string${build([defaultForeground])}';
    }
  }

  /// Joins [params] into a full ANSI escape sequence.
  @pragma('vm:prefer-inline')
  static String build(List<String> params) => escapePrefix + params.join(';') + escapeSuffix;

  /// The ANSI escape code prefix, used to start an ANSI parameter sequence.
  static const escapePrefix = '\x1b[';

  /// The ANSI escape code suffix, used to return to normal text output.
  static const escapeSuffix = 'm';

  /// The ANSI escape param to disable bold (the default).
  static const notBold = '22';

  /// The ANSI escape param to disable underline (the default).
  static const notUnderlined = '24';

  /// The ANSI escape param to disable blinking (the default).
  static const notBlinking = '25';

  /// The ANSI escape param to disable inverse (the default).
  static const notInverse = '27';

  /// The ANSI escape param to disable hidden (the default).
  static const notHidden = '28';

  /// The ANSI escape param to reset foreground color.
  static const defaultForeground = '39';

  /// The ANSI escape param to reset background color.
  static const defaultBackground = '49';

  /// Regular Expression to match ansi codes.
  static const regexp = r'\x1b\[[0-9;]+m';

  /// Compiled Regular Expression to match ansi codes.
  static final regex = RegExp(regexp);
}
