import 'dart:async';
import 'dart:collection';

import 'package:meta/meta.dart';
import 'package:rxdart/rxdart.dart' show Subject;
import 'package:tuple/tuple.dart';

/// A stream that queues events when it does not have a listener to accept them.
///
/// The number of listeners is tracked via onListen and onCancel methods, and when it is 0 events are sent to the queue.
/// 
/// To maintain the order of events added, events are only added to the stream if the queue is empty. This ensure that, for example, when a non-sync [PatientStreamQueue] is performing a flush after receiving a listener, events added before the flush has completed will not appear in the stream before events already in the queue.
class PatientStreamQueue<E> extends Subject<E> {
  /// Creates a new instance of [PatientStreamQueue].
  /// 
  /// In addition to being passed on to the underlying stream controller, when [sync] is false, the flush performed on listen will be done so asynchronously. Within the while loop that continues until the queue is empty, the code that adds the queue element to the stream will be wrapped in a [Future], causing it to be executed in a later microtask. 
  // [sync] may be enabled to ensure that when listeners are present, events are emitted in the order that they are added. This can be used to enure that events intended to be logged show up before a subsequent call to io.stdin is used to read a user's response.
  PatientStreamQueue({
    bool sync = false,
  }) : this.internal(StreamController<E>.broadcast(sync: sync), sync: sync);

  /// When specified, [internalController] **must** be a broadcast stream.
  PatientStreamQueue.from(StreamController<E> internalController, {required bool sync}) : this.internal(internalController, sync: sync);

  /// [internalController] **must** be a broadcast stream.
  PatientStreamQueue.internal(this.internalController, {required this.sync})
      : _queue = Queue(),
        broadcastStream = internalController.stream.asBroadcastStream(),
        // sink = internalController.sink,
        super(internalController, internalController.stream) {
    internalController.onListen = _onListen;
    internalController.onCancel = _onCancel;
    if (internalController.hasListener) _onListen();
  }

  /// The internal [StreamController] of this [PatientStreamQueue]'s [Stream]
  @protected
  final StreamController<E> internalController;

  /// This [PatientStreamQueue]'s broadcast stream.
  final Stream<E> broadcastStream;

  final Queue<dynamic> _queue;

  int _listeners = 0;

  final bool sync;

  @override
  void add(E event) {
    // if no listeners are registered or the queue has not finished being flushed, skip the stream and add to the queue.
    if (_listeners == 0 || _queue.isNotEmpty) {
      _queue.add(event);
      return;
    }

    super.add(event);
  }

  // @override
  // void onAdd(E event) {
  //   // in addition to pushing an event to the stream, if no listeners are active, add the event to the queue to be scooped up later.
  //   if (_listeners == 0 || _queue.isNotEmpty) _queue.add(event);
  // }

  @override
  void addError(Object error, [StackTrace? stackTrace]) {
    // if no listeners are registered or the queue has not finished being flushed, skip the stream and add to the queue.
    if (_listeners == 0 || _queue.isNotEmpty) {
      _queue.add(Tuple2(error, stackTrace));
      return;
    }

    if (stackTrace != null) {
      super.addError(error, stackTrace);
    } else {
      super.addError(error);
    }
  }

  // @override
  // void onAddError(Object error, [StackTrace? stackTrace]) {
  //   if (_listeners == 0 || _queue.isNotEmpty) _queue.add(Tuple2(error, stackTrace));
  // }

  /// Disposes of the [PatientStreamQueue] and [internalController], waiting for any events to fire.
  @override
  Future<void> close({bool wait = true}) async {
    if (wait && _queue.isNotEmpty) {
      // it is assumed that after _listened is complete, the queue will be empty.
      await _flushCompleter!.future;
    }

    await kill();
  }

  /// Kills the [PatientStreamQueue] and [internalController].
  ///
  /// The returned future will complete once [internalController.close] returns, and the future will pass along any encountered error.
  Future<void> kill() async {
    await internalController.close();
  }

  Completer<void>? _flushCompleter;

  /// Future that will complete when the asynchronous buffer flush triggered by a call to [listen] has finished.
  Future<void>? get flushed => _flushCompleter?.future;

  void _onListen() {
    // if _listened is still being completed, a listener is still being setup, and the queue is still being flushed.
    if (_flushCompleter != null && !_flushCompleter!.isCompleted) {
      throw StateError('_onListen called while another listener was being setup.');
    }

    // if _listened is complete, the listener is finished being setup, and the queue has been flushed to the stream, where the listener will scoop up the events.
    _flushCompleter = Completer<void>();
    _listeners++;

    if (sync) {
      _flushSync();
      _flushCompleter!.complete();
    } else {
      _flush().whenComplete(() {
        _flushCompleter!.complete();
      });
    }
  }

  void _flushSync() {
    while (_queue.isNotEmpty) {
      // if (internalController.isClosed) {
      //   break;
      // }
      final current = _queue.removeFirst();
      if (current is Tuple2<Object, StackTrace>) {
        internalController.addError(current.item1, current.item2);
      } else {
        internalController.add(current);
      }
    }
  }

  Future<void> _flush() async {
    while (_queue.isNotEmpty) {
      final cont = await Future<bool>(() {
        if (internalController.isClosed) {
          return false;
        }
        final current = _queue.removeFirst();
        if (current is Tuple2<Object, StackTrace>) {
          internalController.addError(current.item1, current.item2);
        } else {
          internalController.add(current);
        }

        return true;
      });
      if (!cont) {
        break;
      }
    }
  }

  void _onCancel() {
    _listeners--;
  }
}
