import 'dart:async';
import 'dart:convert';
import 'dart:io';

import 'package:flex/flex.dart';
import 'package:io/io.dart';

typedef FallbackBuilder<T> = T Function();

typedef LineParser<T> = T Function(String line);

/// Prompts the user to enter anything, with the option of not echoing stdin to stdout.
String promptFor({required String prompt, String promptSuffix = ' >>> ', bool echo = true}) {
  stdin.echoMode = echo;

  stdout.write('$prompt$promptSuffix');
  // if this is null, there's really no reasonable way to continue.
  final pwd = stdin.readLineSync()!;
  // Print newline if echo is disabled.
  if (!echo) stdout.writeln();

  stdin.echoMode = true;

  return pwd;
}

/// Prompts the user through the console with [question].
///
/// [fallback] can be used to specify a value of type [T] to be used if the user's response is empty.
///
/// [hint] can be used to override the default hint that will be printed after [question] surrounded by parentheses.
///
/// [parse] must n
T promptSync<T>(String question, {required FallbackBuilder<T> fallback, String? hint, required LineParser<T> parse, bool echo = true, bool debug = false}) {
  // this may throw an error. we can't handle it because it could be a security issue for the input to be echoed.
  stdin.echoMode = echo;

  stdout.write('$question ${hint != null ? '($hint)' : ''} >>> ');

  // we can't reasonably handle a null input.
  final line = stdin.readLineSync()!;

  // Print newline if echo is disabled.
  if (!echo) stdout.writeln();

  stdin.echoMode = true;

  if (debug) {
    print('INPUT = "$line"');
  }

  if (line.isEmpty) {
    return fallback();
  }
  
  return parse(line);
}

/// Prompts the user through the console with [question] and returns their yes or no response.
//ignore: avoid_positional_boolean_parameters
bool shouldPromptSync(String question, [bool def = true]) => promptSync<bool>(question, fallback: () => def, hint: '${def ? 'Y' : 'y'}/${def ? 'n' : 'N'}', parse: formsOfYes.insensitiveContains);

/// Wrapper around [sharedStdin] providing each character as a [String]
final Stream<String> stdinUtf8 = sharedStdIn.transform(utf8.decoder).asBroadcastStream();

/// Wrapper around [stdinUtf8] splitting at line endings
final Stream<String> stdinLined = stdinUtf8.transform(const LineSplitter());

/// Prompts the user through the console with [question].
///
/// [fallback] can be used to specify a value of type [T] to be used if the user's response is empty.
///
/// [hint] can be used to override the default hint that will be printed after [question] surrounded by parentheses.
///
/// [parse] must n
Future<T> prompt<T>(String question, {required FallbackBuilder<FutureOr<T>> fallback, String? hint, required LineParser<FutureOr<T>> parse, bool echo = true, bool debug = false}) async {
  // this may throw an error. we can't handle it because it could be a security issue for the input to be echoed.
  stdin.echoMode = echo;

  stdout.write('$question${hint != null ? ' ($hint)' : ''} >>> ');

  final line = await stdinLined.first;

  // Print newline if echo is disabled.
  if (!echo) stdout.writeln();

  stdin.echoMode = true;

  if (debug) {
    print('INPUT = "$line"');
  }

  if (line.isEmpty) {
    return await fallback();
  }
  
  return await parse(line);
}

/// Prompts the user through the console with [question] and returns their yes or no response.
//ignore: avoid_positional_boolean_parameters
Future<bool> shouldPrompt(String question, [bool def = true]) => prompt<bool>(question, fallback: () => def, hint: '${def ? 'Y' : 'y'}/${def ? 'n' : 'N'}', parse: formsOfYes.insensitiveContains);

/// Various ridiculous letter combinations that are generally accepted as "yes" in English.
const formsOfYes = {
  'yup',
  'yes',
  'sure',
  'ye',
  'yeet',
  'yet',
  'yep',
  'yop',
  'why not',
  'y not',
  'y',
  'you betcha',
  'you becha',
};
