import 'dart:async';
import 'dart:collection';

import 'package:collection/collection.dart';
import 'package:flex/flex.dart';

import 'test_core.dart';

/// Buffer of lines of text.
class LineBuffer {
  final Queue<String> lines = Queue();

  /// Adds a line to the buffer.
  void write(String line) => lines.add(line);
}

Future<void> main() async {
  final eq = const ListEquality().equals;

  await Runner([
    Group(
      id: 'patient_stream_queue',
      description: 'Tests relating to [PatientStreamQueue].',
      children: [
        Test(
          id: 'waiting_nosync',
          description: 'Testing the behaviour of a [PatientStreamQueue] without sync enabled when it is closed with waiting enabled.',
          function: (pause, resume) async {
            try {
              final buf = LineBuffer();

              final waiting = PatientStreamQueue<String>(sync: false);
              waiting.add('Hello!');
              waiting.add('How are you?');
              waiting.add('> I am good.');
              waiting.listen(buf.write);
              // await waiting.
              waiting.add('That is good!');
              await waiting.close(wait: true);

              assert(buf.lines.first == 'Hello!', 'First element in buffer must be the first element added to the stream.');
              assert(buf.lines.last == 'That is good!', 'Last element in buffer must be the element sent after the listener is setup.');
              assert(buf.lines.length == 4, 'The length of the buffer must be the number of elements added.');
              assert(
                eq(
                  buf.lines.toList(growable: false),
                  [
                    'Hello!',
                    'How are you?',
                    '> I am good.',
                    'That is good!',
                  ],
                ),
                'The elements of the buffer as a list must match a list of all the elements added to the stream.',
              );
            } on AssertionError catch (e) {
              return AssertionTestResult('waiting_nosync', e);
            } catch (e, st) {
              return ErrorTestResult('waiting_nosync', e, st);
            }

            return const SuccessTestResult('waiting_nosync');
          },
        ),
        Test(
          id: 'waiting_sync',
          description: 'Testing the behaviour of a [PatientStreamQueue] with sync enabled when it is closed with waiting enabled.',
          function: (pause, resume) async {
            try {
              final buf = LineBuffer();

              final waiting = PatientStreamQueue<String>(sync: true);
              waiting.add('Hello!');
              waiting.add('How are you?');
              waiting.add('> I am good.');
              waiting.listen(buf.write);
              waiting.add('That is good!');
              await waiting.close(wait: true);

              assert(buf.lines.first == 'Hello!', 'First element in buffer must be the first element added to the stream.');
              assert(buf.lines.last == 'That is good!', 'Last element in buffer must be the element sent after the listener is setup.');
              assert(buf.lines.length == 4, 'The length of the buffer must be the number of elements added.');
              assert(
                eq(
                  buf.lines.toList(growable: false),
                  [
                    'Hello!',
                    'How are you?',
                    '> I am good.',
                    'That is good!',
                  ],
                ),
                'The elements of the buffer as a list must match a list of all the elements added to the stream.',
              );
            } on AssertionError catch (e) {
              return AssertionTestResult('waiting_sync', e);
            } catch (e, st) {
              return ErrorTestResult('waiting_sync', e, st);
            }

            return const SuccessTestResult('waiting_sync');
          },
        ),
        Test(
          id: 'nonwaiting_nosync',
          description: 'Testing the behaviour of a [PatientStreamQueue] without sync enabled when it is closed with waiting disabled.',
          function: (pause, resume) async {
            try {
              // await runZoned<Future<void>>(
              //   () async {
                  final buf = LineBuffer();

                  final waiting = PatientStreamQueue<String>(sync: false);
                  waiting.add('Hello!');
                  waiting.add('How are you?');

                  for (var i = 0; i < 10000; i++) {
                    await Future(() => waiting.add('> I am good.'));
                  }

                  waiting.listen(buf.write);
                  waiting.add('That is good!');
                  await waiting.close(wait: false);

                  assert(buf.lines.isEmpty, 'When closed with wait=false immediately after adding one or more new events to the queue, adding a new listener, and adding one or more new events to the queue, a [PatientStreamQueue] should have a length of 0, because the close should complete before the onListen-triggered flush.');
              //   },
              //   zoneSpecification: const ZoneSpecification(),
              // );
            } on AssertionError catch (e) {
              return AssertionTestResult('nonwaiting_nosync', e);
            } catch (e, st) {
              return ErrorTestResult('nonwaiting_nosync', e, st);
            }

            return const SuccessTestResult('nonwaiting_nosync');
          },
        ),
        Test(
          id: 'nonwaiting_sync',
          description: 'Testing the behaviour of a [PatientStreamQueue] with sync enabled when it is closed with waiting disabled.',
          function: (pause, resume) async {
            try {
              final buf = LineBuffer();

              final waiting = PatientStreamQueue<String>(sync: true);
              waiting.add('Hello!');
              waiting.add('How are you?');

              for (var i = 0; i < 10000; i++) {
                await Future(() => waiting.add('> I am good.'));
              }

              waiting.listen(buf.write);
              waiting.add('That is good!');
              await waiting.close(wait: false);

              assert(buf.lines.first == 'Hello!', 'First element in buffer must be the first element added to the stream.');
              assert(buf.lines.last == 'That is good!', 'Last element in buffer must be the element sent after the listener is setup.');
              assert(buf.lines.length == 10000 + 3, 'The length of the buffer must be the number of elements added.');
              // assert(
              //   eq(
              //     buf.lines.toList(growable: false),
              //     [
              //       'Hello!',
              //       'How are you?',
              //       '> I am good.',
              //       'That is good!',
              //     ],
              //   ),
              //   'The elements of the buffer as a list must match a list of all the elements added to the stream.',
              // );
            } on AssertionError catch (e) {
              return AssertionTestResult('nonwaiting_nosync', e);
            } catch (e, st) {
              return ErrorTestResult('nonwaiting_nosync', e, st);
            }

            return const SuccessTestResult('nonwaiting_nosync');
          },
        ),
      ],
    ),
  ]).test();
}
