import 'dart:async';
import 'dart:io' as io;
import 'dart:math' as math;

import 'package:flex/expose_flutter.dart';
import 'package:flex/flex.dart' show AnsiColor;
import 'package:meta/meta.dart' show protected;
import 'package:stack_trace/stack_trace.dart';

// import 'package:test/test.dart' as dtest;

// cant be exported because dart(export_legacy_symbol) is a compile time error.
// export 'package:test/test.dart' hide test, group;

/// The outcome of a test, indicating whether it was successful.
enum TestOutcome {
  /// Indicates that a test completed successfully.
  success,

  /// Indicates that a test failed due to an [AssertionError].
  assertion,

  /// Indicates that a test failed due to an unknown [Error] or [Exception].
  error,
}

abstract class TestResult {
  const TestResult._(this.testId);

  /// The test id of the test
  final String testId;
}

class SuccessTestResult extends TestResult {
  const SuccessTestResult(String testId) : super._(testId);
}

class ErrorTestResult extends TestResult {
  const ErrorTestResult(String testId, this.error, this.stackTrace) : super._(testId);

  /// Error that caused the test to fail.
  final Object error;

  /// StackTrace generated catching the error that caused the test to fail.
  final StackTrace stackTrace;
}

class AssertionTestResult extends TestResult {
  const AssertionTestResult(String testId, this.assertion) : super._(testId);

  /// Error thrown by the failed [assert].
  final AssertionError assertion;
}

class InheritedTestResult extends TestResult {
  const InheritedTestResult(String testId) : super._(testId);
}

class WorkingTestResult extends TestResult {
  const WorkingTestResult(String testId) : super._(testId);
}

typedef TestAction = void Function();

typedef TestFunction = Future<TestResult> Function(TestAction pause, TestAction resume);

// typedef TestFailer = void Function(TestResult result);

typedef TestLogger = int Function(List<String> prefixes, TestResult result, Duration time, {int erase});

/// A class implementing testable will be timed as a test, and generally will follow the behaviour of a test.
abstract class Testable {
  Testable(this.id, this.description);

  /// Unique but descriptive tag to identify the test.
  final String id;

  /// Short description of the test, similar to an assertion message.
  final String description;

  /// Test stopwatch, used to time the test(s).
  @protected
  final Stopwatch stopwatch = Stopwatch();

  /// The last element of the returned test results is assumed to be the result of this testable.
  Future<List<TestResult>> test(TestAction pause, TestAction resume, List<String> ancestors, TestLogger log);
}

// abstract class FailTarget implements Testable {
//   /// Mark this fail target a failure with [result].
//   void fail(TestResult result);
// }

/// A test. The test will be marked a failure if and only if [function] fails. Failures in [dependents] have no affect on the outcome of the test. Failures in [dependents] will, however, bubble up to groups, marking them failures.
class Test extends Testable {
  Test({required String id, required String description, required this.function, this.dependents = const []}) : super(id, description);

  /// Body of the test. This is the code that will be timed.
  final TestFunction function;

  final List<Testable> dependents;

  @override
  Future<List<TestResult>> test(TestAction pause, TestAction resume, List<String> ancestors, TestLogger log) async {
    final List<TestResult> results = [];

    void mPause() {
      pause();
      stopwatch.stop();
    }

    void mResume() {
      stopwatch.start();
      resume();
    }

    final mAncest = [...ancestors, id];

    bool done = false;
    // this is defined as a function to minimize the stuff happening during the stopwatch time.
    void setDone() => done = true;

    mResume();
    final fut = function(mPause, mResume)..whenComplete(setDone);
    final len = log(mAncest, WorkingTestResult(id), stopwatch.elapsed);
    // while it's not done, spam the console with the elapsed time.
    while (!done) {
      // seems that due to Dart's sync vs. async system, a sync call in an async function is kind of still sync, so this while loop with no async in the body will be executed synchronously, never giving the test time to complete (it seems). if that is true, it is strange that the stopwatch's elapsed time is increasing; of course, it is possible that it is basing elapsed time off of like, the started time minus now or whatever.
      // await Future.delayed(const Duration(microseconds: 1));
      await Future(() => log(mAncest, WorkingTestResult(id), stopwatch.elapsed, erase: len));
      // log(mAncest, WorkingTestResult(id), stopwatch.elapsed, first: false);
    }
    // at this point, it's done anyway, but whatever.
    final res = await fut;
    mPause();

    log(mAncest, res, stopwatch.elapsed, erase: len);

    if (res is SuccessTestResult) {
      for (final dependent in dependents) {
        results.addAll(await dependent.test(mPause, mResume, mAncest, log));
      }
    }

    results.add(res);

    return results;
  }
}

/// A group of tests. The group will be marked a failure if any one of its tests fail. Note, however, that tests in a group are not depenedent upon eachother, hence if a test in a group fails, it will not abort the execution of subsequent tests. For this behaviour, make a test a child of another test.
class Group extends Testable {
  Group({required String id, required String description, required this.children}) : super(id, description);

  final List<Testable> children;

  @override
  Future<List<TestResult>> test(TestAction pause, TestAction resume, List<String> ancestors, TestLogger log) async {
    final List<TestResult> results = [];

    void mPause() {
      stopwatch.stop();
      pause();
    }

    void mResume() {
      resume();
      stopwatch.start();
    }

    final mAncest = [...ancestors, id];

    for (final child in children) {
      results.addAll(await child.test(mPause, mResume, mAncest, log));
    }

    if (results.any((element) => element is! SuccessTestResult)) {
      results.add(InheritedTestResult(id));
    } else {
      results.add(SuccessTestResult(id));
    }

    log(mAncest, results.last, stopwatch.elapsed);

    return results;
  }
}

class Runner {
  Runner(this.children);

  final List<Testable> children;

  final Stopwatch stopwatch = Stopwatch();

  String logStatusColor(TestResult result) {
    const errColor = Color(0xFFf54242);
    const sucColor = Color(0xFF42f59b);
    const warnColor = Color(0xFFf5b042);
    // const infoColor = Color(0xFF5a42f5);
    // const sucAltColor = Color(0xFFc8f542);

    if (result is WorkingTestResult) {
      return warnColor.ansiWrap('Working');
    } else if (result is SuccessTestResult) {
      return sucColor.ansiWrap('Success');
    } else if (result is AssertionTestResult) {
      return errColor.ansiWrap('Assert \nAssertion failed: ${result.assertion.message}');
    } else if (result is ErrorTestResult) {
      return errColor.ansiWrap('Error  \nError occurred: ${result.error}');
    } else if (result is InheritedTestResult) {
      return errColor.ansiWrap('Error  \nError occurred in child test');
    } else {
      throw UnsupportedError('Unknown test result type ${result.runtimeType}.');
    }
  }

  String logStatus(String status, Duration time) => AnsiColor.build(['1']) + time.toString() + AnsiColor.build([AnsiColor.notBold]) + ' - $status';

  int log(List<String> prefixes, TestResult result, Duration time, {int erase = 0}) {
    String status = logStatus(logStatusColor(result), time);
    if (result is SuccessTestResult || result is WorkingTestResult || result is InheritedTestResult) {
    } else if (result is AssertionTestResult) {
      final st = result.assertion.stackTrace;
      if (st != null) {
        status += '\n' + Trace.from(st).terse.toString();
      }
    } else if (result is ErrorTestResult) {
      status += '\n' + Trace.from(result.stackTrace).terse.toString();
    } else {
      throw UnsupportedError('Unknown test result type ${result.runtimeType}.');
    }
    // erase current line.
    // final statusLines = status.split('\n').length;
    // for (int i = 0; i < statusLines; i++) {
    // clear line (mode 1), move to column (col 0).
    // io.stdout.write('\x1b[1K\x1b[0G');
    // }

    if (result is InheritedTestResult) {
      // breakpoint;
      (){}();
    }

    // length of the first line of status, stripped of ansi codes.
    final len = status.replaceAll(AnsiColor.regex, '').split('\n')[0].length;

    if (erase == 0) {
      io.stdout.write('${prefixes.join(' -> ')}: '.padRight(io.stdout.terminalColumns - len-1) + status + (result is WorkingTestResult ? '' : '\n'));
    } else {
      // move cursor back the magic number of 24 characters. this is based on the length of the status string when the status is "0:00:00.000000 - Waiting"
      // const magicNumber = 24;
      io.stdout.write('\x1b[${len}D');
      // io.stdout.write('\x1b[${53}D${status.split('\n')[0].length}');
      // if (status.length < 24) {
      // io.stdout.write(status.padLeft(24 - status.length) + (result is WorkingTestResult ? '' : '\n'));
      // } else {
      // if it is working, don't print newline. otherwise, do.
      io.stdout.write(status.padLeft(erase) + (result is WorkingTestResult ? '' : '\n'));
      // }
    }

    return  status.length;
  }

  void pause() {
    stopwatch.stop();
  }

  void resume() {
    stopwatch.start();
  }

  Future<List<TestResult>> test() async {
    final List<TestResult> results = [];

    for (final child in children) {
      results.addAll(await child.test(pause, resume, const ['runner'], log));
    }

    if (results.any((element) => element is! SuccessTestResult)) {
      results.add(const InheritedTestResult('runner'));
    } else {
      results.add(const SuccessTestResult('runner'));
    }

    log(const ['runner'], results.last, stopwatch.elapsed);

    return results;
  }
}
